type 'a abin =
	  Vide
	| N of 'a abin * 'a * 'a abin
;;

(* Question 1 *)

let abin_exemple = 
	N(
		N(
			Vide,
			4,
			N(Vide,0,Vide)
		),
		5,
		N(
			Vide,
			8,
			N(
				N(Vide,3,Vide),
				1,
				N(Vide,5,Vide)
			)
		)
	)
;;

(* Question 2 *)

let rec nb_feuilles a = match a with
	| Vide 			-> 0
	| N(Vide,_,Vide) 	-> 1
	| N(g,_,d) 		-> (nb_feuilles g) + (nb_feuilles d)
;;

let rec taille a = match a with
	| Vide		-> 0
	| N(g,_,d)	-> (taille g) + (taille d) + 1
;;

let rec hauteur a = match a with
	| Vide		-> -1
	| N(g,_,d)	-> 1 + (max (hauteur g) (hauteur d))
;;

nb_feuilles abin_exemple ;; (* doit renvoyer 3 *)
taille abin_exemple ;;      (* doit renvoyer 7 *)
hauteur abin_exemple ;;     (* doit renvoyer 3 *)

(* Question 3 *)

let rec parcours_infixe a = match a with
	| Vide 		-> []
	| N(g,e,d) 	-> (parcours_infixe g)@(e::(parcours_infixe d))
;;

let rec parcours_prefixe a = match a with
	| Vide		-> []
	| N(g,e,d)	-> e::((parcours_prefixe g)@(parcours_prefixe d))
;;

let rec parcours_postfixe a = match a with
	| Vide		-> []
	| N(g,e,d)	-> (parcours_postfixe g)@(parcours_postfixe d)@[e]
;;

parcours_infixe abin_exemple ;;   (* doit renvoyer [4; 0; 5; 8; 3; 1; 5] *)
parcours_prefixe abin_exemple ;;  (* doit renvoyer [5; 4; 0; 8; 1; 3; 5] *)
parcours_postfixe abin_exemple ;;(* doit renvoyer [0; 4; 3; 5; 1; 8; 5] *)

(* Question 4 *)


let est_ABR a =
 let rec est_croissante l =
   match l with
    | [] -> true
    | [x] -> true
    | x :: y :: q ->  (x <= y) && (est_croissante (y::q))
 in est_croissante (parcours_infixe a);;

est_ABR abin_exemple ;;  (* doit renvoyer false *)


(* Question 5 *)

let rec somme a = match a with
	| Vide 		-> 0
	| N(g,e,d)	-> (somme g) + e + (somme d)
;;

somme abin_exemple ;; (* doit renvoyer 26 *)

(* Question 6 *)

let appartient (a,b) = 
	let rec aux a = match a with
	| Vide 		-> false
	| N(g,e,d)	-> (e == b) || (aux g) || (aux d)
	in aux a
;;

(* Question 7 *)

let rec appartient_et_compte (a, b) = match a with
  | Vide -> (false, 0)
  | N (g, x, d) -> let (bg, cg) = appartient_et_compte (g, b) and
     (bd, cd) = appartient_et_compte (d, b) in
     if x = b then (true, cg + cd + 1)
     else (bg || bd, cg + cd)
;;

appartient (abin_exemple,8);; (* doit renvoyer true *)
appartient (abin_exemple,6);; (* doit renvoyer false *)

appartient_et_compte (abin_exemple,5);; (* doit renvoyer (true,2) *)
appartient_et_compte (abin_exemple,6);; (* doit renvoyer (false,0) *)


(* Question 8 *)

let parents a b =
	let rec aux ab p = match ab with
		| Vide 		-> []
		| N(g,e,d) 	when e == b
				-> p::((aux g e)@(aux d e))
		| N(g,e,d) 	-> (aux g e)@(aux d e)
	in match a with
	| Vide -> []
	| N(g,e,d) -> (aux g e)@(aux d e)
;;

parents abin_exemple 5;; (* doit renvoyer [1] *)
parents abin_exemple 1;; (* doit renvoyer [8] *)
parents abin_exemple 6;; (* doit renvoyer [] *)

(* Question 9 *)

let exemple_peres = [|4;8;1;1;5;-1;4;4;5|] ;;

let exemple_fils = [|[ ];[3;2];[ ];[ ];[6;0;7];[4;8];[ ];[ ];[1]|] ;;

(* Question 10 *)

let racine ap =
 let rec aux i =
  if ap.(i) = - 1 then i else aux ap.(i)
 in aux 0
;;

racine exemple_peres ;; (* doit renvoyer 5 *)

(* Question 11 *)

let hauteur_par_peres ap i =
	let rec aux i =
		if ap.(i) == -1 then 0
		else 1 + (aux (ap.(i)))
	in aux i
;;

hauteur_par_peres exemple_peres 7 ;; (* doit renvoyer 2 *)

(* Question 12 *)

let peres2files af =
	let ap = Array.make (Array.length af) (-1) in
	let rec aux lfils pere = match lfils with
	| [] 		-> ()
	| fils::suite 	-> begin ap.(fils) <- pere ;
				 aux suite pere end
	in for i=0 to (Array.length af)-1 do
		aux (af.(i)) i
	done ; ap
;;

peres2files exemple_fils ;; (* doit renvoyer [|4; 8; 1; 1; 5; -1; 4; 4; 5|] *)

(* Question 13 *)

(* Note : on suppose ici que l'arbre binaire passé en
   paramètre est convertible en un tableau des fils,
   c'est à dire que ses étiquettes sont bien dans [1,n]
   et que chaque étiquette n'apparait qu'une fois. *)

let abin2fils a =
	let etiq a = match a with
		| Vide -> failwith "cas non pris en charge"
		| N(_,e,_) -> e
	in
	let t = taille a in
	let afils = Array.make t [] in
	let rec transform a = match a with
		| Vide 			-> ()
		| N(Vide,p,Vide) 	-> ()
		| N(g,p,Vide) 	-> (afils.(p) <- (etiq g)::(afils.(p)); transform g)
		| N(Vide,p,d) 	-> (afils.(p) <- (etiq d)::(afils.(p)); transform d)
		| N(g,p,d)		-> (afils.(p) <- (etiq g)::((etiq d)::(afils.(p))); transform g; transform d)
	in transform a; afils
;;

let abin_exemple2 = 
	N(
		N(
			Vide,
			4,
			N(Vide,0,Vide)
		),
		5,
		N(
			Vide,
			2,
			N(
				N(Vide,3,Vide),
				1,
				N(Vide,6,Vide)
			)
		)
	)
;;


abin2fils abin_exemple2;;



(* Question 15 *)

type file =  (int abin) list ref;;

let enfile a (f: file) = f := !f @ [a];;

let estVide f = !f = [];;

let defile f = match !f with
  | [] -> failwith "la liste est vide"
  | x :: q -> f := q; x;;

let parcours_largeur a =
	let resultat = ref [] 
	and f = ref []
	in begin
		enfile a f;
		while not (estVide f) do
			let ab = defile f in
			match ab with
			| N(g,e,d) -> begin
				resultat := e::(!resultat);
				enfile g f;
				enfile d f
				end
			| _ -> ()
		done ;
		List.rev (!resultat)
	end
;;

parcours_largeur abin_exemple ;; (* doit renvoyer [5; 4; 8; 0; 1; 3; 5] *)



